cimport cython
cimport cpython
from libc.math cimport log, exp, fabs, sqrt, cos, sin, tan, asin, atan, atan2, M_PI

from _helpers cimport unpack_point, to_radians, to_degrees

from geo.constants import (
    EARTH_MEAN_RADIUS as py_EARTH_MEAN_RADIUS,
    EARTH_MEAN_DIAMETER as py_EARTH_MEAN_DIAMETER,
    EARTH_EQUATORIAL_RADIUS as py_EARTH_EQUATORIAL_RADIUS,
    EARTH_EQUATORIAL_METERS_PER_DEGREE as py_EARTH_EQUATORIAL_METERS_PER_DEGREE,
    I_EARTH_EQUATORIAL_METERS_PER_DEGREE as py_I_EARTH_EQUATORIAL_METERS_PER_DEGREE
)

cdef double EARTH_MEAN_RADIUS = py_EARTH_MEAN_RADIUS
cdef double EARTH_MEAN_DIAMETER = py_EARTH_MEAN_DIAMETER
cdef double EARTH_EQUATORIAL_RADIUS = py_EARTH_EQUATORIAL_RADIUS
cdef double EARTH_EQUATORIAL_METERS_PER_DEGREE = py_EARTH_EQUATORIAL_METERS_PER_DEGREE
cdef double I_EARTH_EQUATORIAL_METERS_PER_DEGREE = py_I_EARTH_EQUATORIAL_METERS_PER_DEGREE
cdef HALF_PI = M_PI / 2.0

def _approximate_distance(point1, point2):
    cdef double lon1, lat1, lon2, lat2

    unpack_point(point1, &lon1, &lat1)
    unpack_point(point2, &lon2, &lat2)

    cdef double cos_lat
    cos_lat = cos((lat1+lat2)/2.0)

    cdef double dx = (lat2 - lat1)
    cdef double dy = (cos_lat*(lon2 - lon1))

    return EARTH_MEAN_RADIUS*sqrt(dx**2 + dy**2)


def _haversine_distance(point1, point2):
    cdef double lon1, lat1, lon2, lat2

    unpack_point(point1, &lon1, &lat1)
    unpack_point(point2, &lon2, &lat2)

    cdef double dlat = (lat2 - lat1)
    cdef double dlon = (lon2 - lon1)
    cdef double a = (
        sin(dlat * 0.5)**2 +
        cos(lat1) * cos(lat2) * sin(dlon * 0.5)**2
    )

    return EARTH_MEAN_DIAMETER * asin(sqrt(a))


def _distance(point1, point2):
    cdef double lon1, lat1, lon2, lat2

    unpack_point(point1, &lon1, &lat1)
    unpack_point(point2, &lon2, &lat2)

    cdef double dlon = fabs(lon1 - lon2)
    cdef double dlat = fabs(lat1 - lat2)

    cdef double numerator = sqrt(
        (cos(lat2)*sin(dlon))**2 +
        ((cos(lat1)*sin(lat2)) - (sin(lat1)*cos(lat2)*cos(dlon)))**2)

    cdef double denominator = (
        (sin(lat1)*sin(lat2)) +
        (cos(lat1)*cos(lat2)*cos(dlon)))

    cdef double c = atan2(numerator, denominator)
    return EARTH_MEAN_RADIUS*c

def _from4326_to3857(point):
    cdef double lon = point[0]
    cdef double lat = point[1]

    cdef double xtile = lon * EARTH_EQUATORIAL_METERS_PER_DEGREE
    cdef double ytile = log(tan(to_radians(45 + lat / 2.0))) * EARTH_EQUATORIAL_RADIUS
    return (xtile, ytile)


def _from3857_to4326(point):
    cdef double x = point[0]
    cdef double y = point[1]
    cdef double lon = x / EARTH_EQUATORIAL_METERS_PER_DEGREE
    cdef double lat = to_degrees(
        2.0 * atan(exp(y/EARTH_EQUATORIAL_RADIUS)) - HALF_PI)
    return (lon, lat)



cimport cython
cimport cpython
from libc.math cimport M_PI


@cython.cdivision(True)
cdef inline double to_degrees(double radians):
    return radians * (180.0 / M_PI)


@cython.cdivision(True)
cdef inline double to_radians(double degrees):
    return (degrees * M_PI) / 180.0


cdef inline unpack_point(point, double *lon, double *lat):
    if (not cpython.PySequence_Check(point)):
        raise TypeError("point must be an iterable of double")
    lon[0] = to_radians(point[0])
    lat[0] = to_radians(point[1])



cimport cython
cimport cpython
from libc.math cimport fabs, sqrt, cos, sin, tan, atan, asin, atan2, log, exp, M_PI
from _helpers cimport unpack_point, to_degrees

from geo.constants import WGS84

cdef double CONVERGENCE_THRESHOLD = 1e-12
cdef int MAX_ITERATIONS = 10
cdef double HALF_PI = M_PI / 2.0
cdef double QUARTER_PI = M_PI / 4.0


def _distance(point1, point2, ellipsoid=WGS84):
    cdef double lon1, lat1, lon2, lat2
    cdef ellipsoid_a = ellipsoid.a
    cdef ellipsoid_b = ellipsoid.b
    cdef ellipsoid_f = ellipsoid.f

    unpack_point(point1, &lon1, &lat1)
    unpack_point(point2, &lon2, &lat2)

    cdef double U1 = atan((1 - ellipsoid_f) * tan(lat1))
    cdef double U2 = atan((1 - ellipsoid_f) * tan(lat2))
    cdef double L = lon2 - lon1

    cdef double sinU1 = sin(U1)
    cdef double cosU1 = cos(U1)
    cdef double sinU2 = sin(U2)
    cdef double cosU2 = cos(U2)

    cdef double sinLambda, sinSigma, sigma, sinAlpha, cosSqAlpha
    cdef double C, Lambda, LambdaPrev

    Lambda = L

    for _ in range(MAX_ITERATIONS):
        sinLambda = sin(Lambda)
        cosLambda = cos(Lambda)
        sinSigma = sqrt(
            (cosU2 * sinLambda) ** 2 +
            (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) ** 2)
        # coincident points
        if sinSigma == 0:
            return 0.0

        cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda
        sigma = atan2(sinSigma, cosSigma)
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha ** 2
        try:
            cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha
        except ZeroDivisionError:
            cos2SigmaM = 0

        C = (ellipsoid_f / 16) * cosSqAlpha * (
            4 + ellipsoid_f * (4 - 3 * cosSqAlpha)
        )
        LambdaPrev = Lambda
        Lambda = (
            L + (1 - C) * ellipsoid.f * sinAlpha * (
                sigma + C * sinSigma * (
                    cos2SigmaM + C * cosSigma * (
                        -1 + 2 * cos2SigmaM ** 2
                    )
                )
            )
        )

        if abs(Lambda - LambdaPrev) < CONVERGENCE_THRESHOLD:
            break
    else:
        # failure to converge
        return None

    cdef double uSq, A, B, deltaSigma, s

    uSq = cosSqAlpha * (ellipsoid_a ** 2 - ellipsoid_b ** 2) / (ellipsoid_b ** 2)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    deltaSigma = B * sinSigma * (cos2SigmaM + B / 4 * (cosSigma *
                 (-1 + 2 * cos2SigmaM ** 2) - B / 6 * cos2SigmaM *
                 (-3 + 4 * sinSigma ** 2) * (-3 + 4 * cos2SigmaM ** 2)))
    s = ellipsoid_b * A * (sigma - deltaSigma)
    return s


def _from4326_to3395(point, ellipsoid=WGS84):
    cdef double lon, lat
    cdef double ellipsoid_a = ellipsoid.a
    cdef double ellipsoid_e = ellipsoid.e

    unpack_point(point, &lon, &lat)

    cdef double e_sin_lat = ellipsoid_e*sin(lat)
    cdef double multiplier1 = tan(QUARTER_PI + lat / 2.0)
    cdef double multiplier2 = pow((1-e_sin_lat)/(1+e_sin_lat), ellipsoid_e/2)
    cdef double E = ellipsoid_a * lon
    cdef double N = ellipsoid_a * log(multiplier1*multiplier2)
    return (E, N)


def _from3395_to4326(point, ellipsoid=WGS84):
    cdef double E, N
    E, N = point
    cdef double a = ellipsoid.a
    cdef double e = ellipsoid.e
    cdef double half_e = e * 0.5

    cdef double m1 = exp(-N / a)
    cdef double m2
    cdef double new_phi
    cdef double phi = HALF_PI - 2.0 * atan(m1)
    cdef double e_sin_phi

    for _ in range(MAX_ITERATIONS):
        e_sin_phi = e*sin(phi)
        m2 = ((1 - e_sin_phi) / (1 + e_sin_phi))**half_e
        new_phi = HALF_PI  - 2.0 * atan(m1 * m2)
        if abs(new_phi - phi) <= CONVERGENCE_THRESHOLD:
            phi = new_phi
            break
        phi = new_phi

    lon = to_degrees(E / a)
    lat = to_degrees(phi)
    return (lon, lat)
